use crate::types::common::DataBlock;
use std::{error::Error, fmt, vec::Vec};

use crate::types::common::convert_from_data_field;
use deku::{deku_derive, DekuReader};

use super::common::{AlphaNumericString, DataGroup};

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct SSVVersionNumber(#[deku(map = "convert_from_data_field")] u16);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct GALSSVersionNumber(#[deku(map = "convert_from_data_field")] u16);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct PSSVersionNumber(#[deku(map = "convert_from_data_field")] u16);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct ConfigurationHeader {
    pub ssv_version: SSVVersionNumber,
    pub galss_version: GALSSVersionNumber,
    pub pss_version: PSSVersionNumber,
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct PCSCReaderName(AlphaNumericString);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct CardType(#[deku(map = "convert_from_data_field")] u8);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct PCSCReader {
    pub name: PCSCReaderName,
    pub card_type: CardType,
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct SESAMVitaleComponentID(#[deku(map = "convert_from_data_field")] u16);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct SESAMVitaleComponentDescription(AlphaNumericString);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct SESAMVitaleComponentVersion(AlphaNumericString);

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct SESAMVitaleComponent {
    pub id: SESAMVitaleComponentID,
    pub description: SESAMVitaleComponentDescription,
    pub version: SESAMVitaleComponentVersion,
}

#[deku_derive(DekuRead)]
#[derive(Debug, PartialEq)]
pub struct ReaderConfiguration {}

#[derive(Debug)]
pub enum ConfigurationError {
    MultipleConfigurationHeaders,
    MissingConfigurationHeader,
}

impl fmt::Display for ConfigurationError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            ConfigurationError::MultipleConfigurationHeaders => {
                write!(f, "Multiple ConfigurationHeader blocks found")
            }
            ConfigurationError::MissingConfigurationHeader => {
                write!(f, "Missing ConfigurationHeader block")
            }
        }
    }
}

impl Error for ConfigurationError {}

#[derive(Debug)]
pub struct Configuration {
    pub configuration_header: ConfigurationHeader,
    pub reader_configurations: Vec<ReaderConfiguration>,
    pub sesam_vitale_components: Vec<SESAMVitaleComponent>,
    pub pcsc_readers: Vec<PCSCReader>,
}

impl TryFrom<Vec<DataBlock>> for Configuration {
    type Error = ConfigurationError;

    fn try_from(data_blocks: Vec<DataBlock>) -> Result<Self, Self::Error> {
        let mut configuration_header: Option<ConfigurationHeader> = None;
        let mut reader_configurations: Vec<ReaderConfiguration> = Vec::new();
        let mut sesam_vitale_components: Vec<SESAMVitaleComponent> = Vec::new();
        let mut pcsc_readers: Vec<PCSCReader> = Vec::new();

        for block in data_blocks {
            match block.inner {
                DataGroup::ConfigurationHeader(header) => {
                    if configuration_header.is_some() {
                        return Err(ConfigurationError::MultipleConfigurationHeaders);
                    }
                    configuration_header = Some(header);
                }
                DataGroup::ReaderConfiguration(configuration) => {
                    reader_configurations.push(configuration)
                }
                DataGroup::SESAMVitaleComponent(component) => {
                    sesam_vitale_components.push(component);
                }
                DataGroup::PCSCReader(reader) => {
                    pcsc_readers.push(reader);
                }
            }
        }
        let configuration_header = match configuration_header {
            Some(header) => header,
            None => return Err(ConfigurationError::MissingConfigurationHeader),
        };

        Ok(Self {
            configuration_header,
            reader_configurations,
            sesam_vitale_components,
            pcsc_readers,
        })
    }
}
